"""
Set of neural network building blocks
"""
from motrack_motion.library.building_blocks.mlp import MLP
from motrack_motion.library.building_blocks.attention import TemporalFirstMultiHeadSelfAttention, TemporalFirstMultiHeadCrossAttention
